from __future__ import annotations
from abc import abstractmethod
import numpy as np
import logging

logger = logging.getLogger('discrete_dists')


Support = tuple[int, int]


class Distribution:
    @abstractmethod
    def probs(self, elements: np.ndarray) -> np.ndarray: ...

    @abstractmethod
    def sample(self, rng: np.random.Generator, n: int) -> np.ndarray: ...

    @abstractmethod
    def stratified_sample(self, rng: np.random.Generator, n: int) -> np.ndarray: ...

    def isr(self, target: Distribution, elements: np.ndarray):
        return target.probs(elements) / self.probs(elements)

    def sample_without_replacement(
        self,
        rng: np.random.Generator,
        n: int,
        attempts: int = 25,
    ) -> np.ndarray:
        elements = self.sample(rng, n)

        # fastpath for the common case that the first sample is already unique
        uniq = set(elements)
        if len(uniq) == n:
            return elements

        for _ in range(attempts):
            needed = n - len(uniq)
            sub = self.sample(rng, 2 * needed)
            uniq |= set(sub)

            if len(uniq) >= n:
                break

        if len(uniq) < n:
            logger.warning(f"Failed to get <{n}> required unique samples. Got <{len(uniq)}>")

        cutoff = min(n, len(uniq))
        out = np.array(list(uniq), dtype=np.int64)[:cutoff]
        return out


    @property
    def is_defunct(self) -> bool:
        return False



import pickle
import numpy as np

from discrete_dists.utils.SumTree import SumTree

class TestSumTree:
    def test_can_add_stuff(self):
        # human readable test
        v = np.array([1, 2, 3, 4, 5])
        truth = 15
        tree = SumTree(500)
        tree.update([0, 5, 8, 102, 358], v)

        assert tree.total() == truth

        # fuzz test
        tree = SumTree(100)
        rng = np.random.default_rng(0)
        for _ in range(100):
            v = rng.standard_normal(100)
            truth = v.sum()
            tree.update(np.arange(100), v)

            assert np.isclose(tree.total(), truth)


    def test_can_sample(self):
        tree = SumTree(50)
        tree.update(np.arange(50), np.ones(50))

        assert tree.total() == 50

        rng = np.random.default_rng(0)
        samples = tree.sample(rng, 10000)

        u, c = np.unique(samples, return_counts=True)
        assert np.all(u == np.arange(50))
        assert np.all(
            (c >= 150) & (c <= 250)
        )

    def test_can_sample_proportionally(self):
        tree = SumTree(10)
        tree.update(np.arange(10), [2**i for i in range(10)])

        # NOTE: it takes a shockingly high number of samples for the proportions
        # to converge even within a single decimal point..
        rng = np.random.default_rng(22)
        samples = tree.sample(rng, 10000000)

        u, c = np.unique(samples, return_counts=True)
        assert np.all(u == np.arange(10))

        for i in range(1, 10):
            assert np.isclose(c[i] / c[i - 1], 2, atol=0.1)

    def test_pickleable(self):
        tree = SumTree(123)
        tree.update(np.arange(123), np.arange(123))

        byt = pickle.dumps(tree)
        tree2 = pickle.loads(byt)

        assert np.all(
            tree.total() == tree2.total()
        )

        tree.update([2], [22])
        tree2.update([2], [22])

        assert np.all(
            tree.total() == tree2.total()
        )

# ----------------
# -- Benchmarks --
# ----------------
class TestBenchmarks:
    def test_sumtree_update(self, benchmark):
        tree = SumTree(100_000)
        rng = np.random.default_rng(0)
        idxs = np.arange(32)
        vals = rng.uniform(0, 2, size=32)

        def _inner(tree: SumTree, idxs, vals):
            tree.update(idxs, vals)

        benchmark(_inner, tree, idxs, vals)

    def test_sumtree_sample(self, benchmark):
        tree = SumTree(100_000)
        rng = np.random.default_rng(0)

        idxs = np.arange(10_000)
        vals = rng.uniform(0, 2, size=10_000)
        tree.update(idxs, vals)

        def _inner(tree: SumTree, rng):
            tree.sample(rng, 32)

        benchmark(_inner, tree, rng)



from dataclasses import dataclass
from typing import Sequence

import numpy as np

from discrete_dists.distribution import Distribution

@dataclass
class SubDistribution:
    """
    A pair joining a distribution with the ratio of
    contribution of this distribution to the mixture.
    """
    d: Distribution
    p: float


class MixtureDistribution(Distribution):
    """
    A mixture over an arbitrary number of sub-distributions.
    Will sample from sub-distributions according to their
    respective probabilities.
    """
    def __init__(self, dists: Sequence[SubDistribution]):
        super().__init__()

        self._dims = len(dists)

        self.dists = [sub.d for sub in dists]
        self._weights = np.array([sub.p for sub in dists])

        assert np.isclose(self._weights.sum(), 1)


    def probs(self, elements: np.ndarray):
        """
        Get the probabilities of the given elements
        under the current distribution.
        """
        sub = np.array([d.probs(elements) for d in self.dists])
        p = self._weights.dot(sub)
        return p

    def sample(self, rng: np.random.Generator, n: int):
        """
        Sample `n` values from the mixture distribution,
        partitioning these `n` values over the various
        sub-distributions according to their respective
        probabilities.
        """
        out = np.empty(n, dtype=np.int64)
        idxs, weights = self.filter_defunct()
        subs = rng.choice(idxs, size=n, replace=True, p=weights)
        elements, counts = np.unique(subs, return_counts=True)

        total = 0
        for element, count in zip(elements, counts, strict=True):
            d = self.dists[int(element)]
            next_t = total + count
            out[total:next_t] = d.sample(rng, count)
            total = next_t

        rng.shuffle(out)
        return out

    def stratified_sample(self, rng: np.random.Generator, n: int):
        """
        Sample `n` values from the mixture distribution,
        partitioning these `n` values over the various
        sub-distributions according to their respective
        probabilities.

        The `m < n` values sampled from each sub-distribution
        will be evenly spaced within that distribution.
        """
        out = np.empty(n, dtype=np.int64)
        idxs, weights = self.filter_defunct()
        subs = rng.choice(idxs, size=n, replace=True, p=weights)
        elements, counts = np.unique(subs, return_counts=True)

        total = 0
        for element, count in zip(elements, counts, strict=True):
            d = self.dists[int(element)]
            next_t = total + count
            out[total:next_t] = d.stratified_sample(rng, count)
            total = next_t

        rng.shuffle(out)
        return out


    def filter_defunct(self):
        """
        Remove any defunct distributions from the mixture, where
        a defunct distribution is defined as having zero support.
        """

        # fastpath for the common case that there are no defunct distributions
        if all(not d.is_defunct for d in self.dists):
            return np.arange(len(self.dists)), self._weights


        dist_idxs = np.array([
            i for i, d in enumerate(self.dists)
            if not d.is_defunct
        ])

        reweighted = self._weights[dist_idxs]
        reweighted /= reweighted.sum()
        return dist_idxs, reweighted



import numpy as np
from typing import Iterable
import discrete_dists.rust as ru

class SumTree(ru.SumTree):
    def __new__(
        cls,
        size: int | None = None,
    ):
        args = [size]
        if args[0] is None:
            args = []

        return super().__new__(cls, *args)

    def __init__(self, size: int):
        super().__init__()

    def update(self, idxs: Iterable[int], values: Iterable[float]):
        a_idxs = np.asarray(idxs, dtype=np.int64)
        a_values = np.asarray(values, dtype=np.float64)

        super().update(a_idxs, a_values)

    def sample(self, rng: np.random.Generator, n: int) -> np.ndarray:
        t = self.total()
        assert t > 0, "Cannot sample when the tree is empty or contains negative values"

        rs = rng.uniform(0, t, size=n)
        return self.query(rs)

    def stratified_sample(self, rng: np.random.Generator, n: int) -> np.ndarray:
        t = self.total()
        assert t > 0, "Cannot sample when the tree is empty or contains negative values"

        buckets = np.linspace(0., 1., n + 1)
        values = np.asarray([
            rng.uniform(buckets[i], buckets[i + 1]) for i in range(n)
        ])

        return self.query(t * values)

    def __getstate__(self):
        return {
            'st': super().__getstate__()
        }

    def __setstate__(self, state):
        super().__setstate__(state['st'])



import pytest
import numpy as np
from discrete_dists.mixture import MixtureDistribution, SubDistribution
from discrete_dists.proportional import Proportional
from discrete_dists.uniform import Uniform


@pytest.fixture
def rng():
    return np.random.default_rng(0)


def test_mixture1(rng):
    m = MixtureDistribution([
        SubDistribution(Uniform(5), p=0.5),
        SubDistribution(Uniform((10, 15)), p=0.5),
    ])

    data = m.sample(rng, 100000)
    vals, counts = np.unique(data, return_counts=True)

    expected_count = 100000 / 10
    thresh = expected_count * 0.05

    support = np.concatenate((np.arange(5), 10 + np.arange(5)))
    assert np.all(vals == support)
    assert np.all(
        counts > (expected_count - thresh)
    ) and np.all(
        counts < (expected_count + thresh)
    )


def test_mixture_defunct(rng):
    m = MixtureDistribution([
        SubDistribution(Uniform(5), p=0.5),
        SubDistribution(Uniform(0), p=0.25),
        SubDistribution(Proportional(10), p=0.25),
    ])

    data = m.sample(rng, 100000)
    vals, counts = np.unique(data, return_counts=True)

    expected_count = 100000 / 5
    thresh = expected_count * 0.05

    assert np.all(vals == np.arange(5))
    assert np.all(
        counts > (expected_count - thresh)
    ) and np.all(
        counts < (expected_count + thresh)
    )



import pytest
import numpy as np
from discrete_dists.uniform import Uniform


@pytest.fixture
def rng():
    return np.random.default_rng(0)


def test_uniform1(rng):
    u = Uniform(10)

    data = u.sample(rng, 1000)
    vals, counts = np.unique(data, return_counts=True)

    assert np.all(vals == np.arange(10))
    assert np.all(counts > 75) and np.all(counts < 125)


def test_uniform2(rng):
    u = Uniform((10, 20))

    data = u.sample(rng, 1000)
    vals, counts = np.unique(data, return_counts=True)

    assert np.all(vals == (10 + np.arange(10)))
    assert np.all(counts > 75) and np.all(counts < 125)


def test_uniform_update(rng):
    u = Uniform(10)
    u.update(np.array([1, 2, 4, 8, 19]))

    data = u.sample(rng, 10000)
    vals, counts = np.unique(data, return_counts=True)

    expected_count = 10000 / 20
    thresh = expected_count * 0.2

    assert np.all(vals == np.arange(20))
    assert np.all(
        counts > (expected_count - thresh)
    ) and np.all(
        counts < (expected_count + thresh)
    )


def test_uniform_sample_wo_replace(rng):
    u = Uniform(10)

    for _ in range(1000):
        data = u.sample_without_replacement(rng, 5)
        assert len(set(data)) == 5


def test_stratified_sample(rng):
    u = Uniform(10)

    data = u.stratified_sample(rng, 5)
    assert len(data) == 5



import numpy as np
from discrete_dists.distribution import Distribution, Support
from discrete_dists.utils.SumTree import SumTree

class Proportional(Distribution):
    """
    A distribution defined over some support
    [lo, hi) where elements within the support
    are sampled proportional to some value.

    By default, the ratios are initially set to 0,
    which inhibits sampling. In order to sample,
    this distribution must first receive updated
    ratios:
    ```python
    p = Proportional(10)
    p.update(
      # elements on the support to update
      elements=np.arange(5),
      # values with which to sample proportionally to
      values=np.array([1, 2, 1, 1, 1]),
    )
    ```
    """
    def __init__(self, support: Support | int):
        if isinstance(support, int):
            support = (0, support)

        self._support = support
        rang = support[1] - support[0]
        self.tree = SumTree(rang)

    # ---------------
    # -- Accessing --
    # ---------------
    def probs(self, elements: np.ndarray) -> np.ndarray:
        """
        Get the probabilities of the given elements
        in the distribution.
        """
        elements = np.asarray(elements)
        elements = elements - self._support[0]

        t = self.tree.total()
        if t == 0:
            return np.zeros(len(elements))

        v = self.tree.get_values(elements)
        return v / t


    @property
    def is_defunct(self) -> bool:
        return self.tree.total() == 0


    # --------------
    # -- Sampling --
    # --------------
    def sample(self, rng: np.random.Generator, n: int) -> np.ndarray:
        """
        Sample `n` values from the distribution. Return
        will be a np.array of integers.
        """
        return self.tree.sample(rng, n) + self._support[0]

    def stratified_sample(self, rng: np.random.Generator, n: int) -> np.ndarray:
        """
        Sample `n` evenly spaced values from the distribution.
        Return will be a np.array of integers.
        """
        return self.tree.stratified_sample(rng, n) + self._support[0]

    # --------------
    # -- Updating --
    # --------------
    def update(self, elements: np.ndarray, values: np.ndarray):
        """
        Update the the proportion values for a given set
        of elements. This changes the shape of the distribution.
        """
        elements = elements - self._support[0]
        self.tree.update(elements, values)

    def update_single(self, element: int, value: float):
        """
        Update the the proportion values for a given single
        element. This changes the shape of the distribution.
        """
        element = element - self._support[0]
        self.tree.update_single(element, value)

    def update_support(self, support: Support | int):
        """
        Shift the entire distribution to be over a new support.
        It is undefined behavior to change the width of the support.
        """
        if isinstance(support, int):
            self._support = (0, support)
        else:
            self._support = support



import pytest
import numpy as np
from discrete_dists.proportional import Proportional


@pytest.fixture
def rng():
    return np.random.default_rng(0)


def test_proportional1(rng):
    p = Proportional(100)
    p.update(
        np.arange(10),
        np.ones(10),
    )

    data = p.sample(rng, 10000)
    vals, counts = np.unique(data, return_counts=True)

    expected_count = 10000 / 10
    thresh = expected_count * 0.05

    assert np.all(vals == np.arange(10))
    assert np.all(
        counts > (expected_count - thresh)
    ) and np.all(
        counts < (expected_count + thresh)
    )


def test_proportional2(rng):
    p = Proportional(100)
    p.update(
        np.arange(10),
        np.ones(10),
    )

    p.update(
        20 + np.arange(10),
        np.ones(10),
    )

    data = p.sample(rng, 100000)
    vals, counts = np.unique(data, return_counts=True)

    expected_count = 100000 / 20
    thresh = expected_count * 0.05

    support = np.concatenate((np.arange(10), 20 + np.arange(10)))
    assert np.all(vals == support)
    assert np.all(
        counts > (expected_count - thresh)
    ) and np.all(
        counts < (expected_count + thresh)
    )


def test_shifted_support(rng):
    p = Proportional((-20, 0))
    p.update(
        np.arange(10) - 10,
        np.ones(10),
    )

    data = p.sample(rng, 10000)
    vals, counts = np.unique(data, return_counts=True)

    expected_count = 10000 / 10
    thresh = expected_count * 0.05

    assert np.all(vals == (np.arange(10) - 10))
    assert np.all(
        counts > (expected_count - thresh)
    ) and np.all(
        counts < (expected_count + thresh)
    )


def test_proportional_stratified1(rng):
    p = Proportional(100)
    p.update(
        np.arange(10),
        np.ones(10),
    )

    p.update(
        20 + np.arange(10),
        np.ones(10),
    )

    data = p.stratified_sample(rng, 5)
    _, counts = np.unique(data, return_counts=True)

    assert np.all(counts == 1)



import numpy as np
import numpy.typing as npt
import discrete_dists.utils.npu as npu
from discrete_dists.distribution import Distribution, Support


class Uniform(Distribution):
    """
    A uniform distribution defined over some support.
    If only an integer support is given, assume support
    is [0, support).
    """
    def __init__(self, support: Support | int):
        if isinstance(support, int):
            support = (0, support)

        self._support: Support = support


    # ------------------------
    # -- Changes to support --
    # ------------------------
    def update(self, elements: np.ndarray) -> None:
        """
        Update the support of the distribution to cover the
        given elements if these are outside the current support.
        """
        self._support = (
            min(self._support[0], elements.min()),
            max(self._support[1], elements.max() + 1)
        )


    def update_single(self, element: int) -> None:
        """
        Update the support of the distribution to cover the
        given element if it is outside the current support.
        """
        self._support = (
            min(self._support[0], element),
            max(self._support[1], element + 1),
        )


    def update_support(self, support: Support | int):
        """
        Update the support of the distribution to the new
        given support.
        """
        if isinstance(support, int):
            self._support = (0, support)
        else:
            self._support = support


    # --------------
    # -- Samplers --
    # --------------
    def sample(self, rng: np.random.Generator, n: int):
        """
        Sample `n` values from the support of the distribution
        with replacement.
        """
        size = self._support[1] - self._support[0]
        if size < 2:
            return self._support[0] + np.zeros(n, dtype=np.int64)

        return rng.integers(*self._support, size=n)


    def stratified_sample(self, rng: np.random.Generator, n: int):
        """
        Sample `n` evenly spaced values from the support
        of the distribution with replacement.
        """
        return npu.stratified_sample_integers(rng, n, *self._support)


    def probs(self, elements: npt.ArrayLike):
        """
        Get the probabilities of the given elements
        under the current distribution.
        """
        d = self._support[1] - self._support[0]
        return np.full_like(elements, fill_value=(1 / d), dtype=np.float64)


    @property
    def is_defunct(self) -> bool:
        return self._support[0] == self._support[1]



import numpy as np

def stratified_sample_integers(rng: np.random.Generator, n: int, lo: int, hi: int):
    buckets = np.linspace(lo, hi, n + 1, endpoint=True, dtype=np.int64)
    samples = [
        rng.integers(buckets[i], buckets[i + 1]) for i in range(n)
    ]
    return np.asarray(samples, dtype=np.int64)



import numpy as np

class SumTree:
    size: int

    def __new__(cls, *args): ...
    def __init__(self, size: int | None = None): ...
    def update(self, idxs: np.ndarray, values: np.ndarray): ...
    def update_single(self, idx: int, value: float): ...
    def get_value(self, idx: int) -> float: ...
    def get_values(self, idxs: np.ndarray) -> np.ndarray: ...
    def total(self) -> float: ...
    def query(self, v: np.ndarray) -> np.ndarray: ...
    def __getstate__(self): ...
    def __setstate__(self, state): ...
